#!/usr/bin/env python
"""
Module PROCESSWRAPPER -- Child Process Wrapper Object
Sub-Package COMM of Package PLIB -- General Python Utilities
Copyright (C) 2008-2015 by Peter A. Donis

This module contains the ``ProcessWrapper`` class, which provides
a portable interface to child processes and their exit status.
This class is used by the specialized forking functions in this
library, and by the ``SigChldMixin`` class in ``plib.comm``.

Note that on Windows the ``multiprocessing`` module is used. (We
purposely do *not* use that module on Unix-type systems, because
it is too heavyweight for what we need here, not to mention that
it doesn't do proper SIGCHLD handling on Unix-type systems--it
doesn't ignore SIGCHLD, as PLIB does by default, so that the
kernel will reap zombies when they arise, but it also does not
trap SIGCHLD and reap zombies itself--instead they are left
hanging around until the parent exits. If there were a lighter
Windows alternative we would use it, but there isn't.)
"""

import os

from ._childwrapper import ChildWrapper

try:
    # Use this to determine what mechanism we'll use to create
    # child processes: either the Unix fork ...
    from os import fork
    OS_FORK = True

except (ImportError, AttributeError):
    # ... or the Windows Rube Goldberg contraption :-)
    OS_FORK = False

if OS_FORK:
    # We have fork (whew!), so do it the right way :-)
    import signal
    
    
    class ProcessWrapper(ChildWrapper):
        # TODO: any way to enforce shutdown_with_parent here?
        
        pid = None
        term_sig = signal.SIGTERM
        
        def start(self):
            if self.pid is None:
                pid = fork()
                if pid == 0:
                    # Child process - note that we reset the termination signal
                    # handler to the default, to ensure that the child is shut
                    # down by it; if more elaborate shutdown signal handling is
                    # desired, the function run in the child must implement it
                    # explicitly. Also, self.term_sig should only be set to a signal
                    # whose default is to shut down the process, unless the child
                    # is going to implement a custom handler that takes care of that.
                    signal.signal(self.term_sig, signal.SIG_DFL)
                    e = self._target(*self._args, **self._kwargs)
                    if not isinstance(e, int):
                        e = 0
                    os._exit(e)
                # Parent process
                self.pid = pid
            else:
                raise RuntimeError("Process already started.")
        
        def stop(self):
            if self.pid is not None:
                os.kill(self.pid, self.term_sig)
        
        def _check_exited(self, wait):
            # Internal method to check exit status
            if self.pid is not None:
                try:
                    pid, status = os.waitpid(self.pid, (os.WNOHANG, 0)[wait])
                except os.error:
                    pid = None
                if pid == self.pid:
                    if os.WIFSIGNALED(status):
                        self._exitcode = -os.WTERMSIG(status)
                    elif os.WIFEXITED(status):
                        self._exitcode = os.WEXITSTATUS(status)
                    else:
                        raise RuntimeError("Invalid process exit status: {}".format(
                                           repr(status)))
                    self.pid = None
        
        def check(self):
            self._check_exited(0)
            return self.pid is None
        
        def wait(self):
            self._check_exited(1)


else:
    # No fork on Windows, have to use the multiprocessing module
    # with some extra gymnastics of our own
    
    from multiprocessing import process
    
    
    class ChildProcess(process.Process):
        
        def _bootstrap(self):
            import sys
            import os
            import itertools
            from multiprocessing import util
            
            try:
                self._children = set()
                self._counter = itertools.count(1)
                try:
                    os.close(sys.stdin.fileno())
                except (OSError, ValueError):
                    pass
                process._current_process = self  # ugly!
                util._finalizer_registry.clear()
                util._run_after_forkers()
                util.info('child process calling self.run()')
                try:
                    # All this just so we can let the run method
                    # return an actual exit code; WTF isn't this
                    # already in there? !@#$%^&*
                    exitcode = self.run()
                    if exitcode is None:
                        exitcode = 0
                finally:
                    util._exit_function()
            except SystemExit, e:
                if not e.args:
                    exitcode = 1
                elif type(e.args[0]) is int:
                    exitcode = e.args[0]
                else:
                    sys.stderr.write(e.args[0] + os.linesep)
                    sys.stderr.flush()
                    exitcode = 1
            except:
                exitcode = 1
                import traceback
                sys.stderr.write('Process {}:{}'.format(self.name, os.linesep))
                sys.stderr.flush()
                traceback.print_exc()
            
            util.info('process exiting with exitcode {:d}'.format(exitcode))
            return exitcode
        
        def run(self):
            # And now we need to actually return the exit code
            # (at least this part is easily implemented...)
            if self._target:
                return self._target(*self._args, **self._kwargs)
    
    
    class ProcessWrapper(ChildWrapper):
        
        process = None
        
        def start(self):
            if self.process is None:
                self.process = ChildProcess(
                    target=self._target, args=self._args, kwargs=self._kwargs)
                self.process.daemon = self.shutdown_with_parent
                self.process.start()
            else:
                raise RuntimeError("Process already started.")
        
        def stop(self):
            self.process.terminate()
        
        def check(self):
            if self.process is None:
                return True
            if not self.process.is_alive():
                self._exitcode = self.process.exitcode
                self.process = None
                return True
            return False
        
        def wait(self):
            self.process.join()
            self._exitcode = self.process.exitcode
            self.process = None



#!/usr/bin/env python
"""
SERVER.PY
Copyright (C) 2008-2015 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

A demonstration server using the self-pipe trick to trap
a shutdown signal and gracefully exit.
"""

import sys
import select
import signal
import socket
from errno import EINTR, ERESTART
from functools import partial
from SocketServer import TCPServer, BaseRequestHandler

from plib.comm.classes import SelfPipe, SigIntMixin

PROCESS = 1
THREAD = 2


class DemoRequestHandler(BaseRequestHandler):
    
    def handle(self):
        print "Got request", self.request, "from", self.client_address, "for", self.server
        sys.stdout.flush()
        sys.stderr.flush()


class DemoServer(SigIntMixin, TCPServer):
    
    allow_reuse_address = True
    
    wrapper_class = pipe = None
    
    def __init__(self, server_address, handler_class, use_pipe, test_race):
        self.__shutdown_flag = False
        if self.wrapper_class:
            self.wrapper_class.shutdown_with_parent = True
        self.use_pipe = use_pipe
        self.test_race = test_race
        self.setup_handlers()
        TCPServer.__init__(self, server_address, handler_class)
    
    def term_sig_handler(self, sig, frame=None):
        print "Terminate signal received by handler"
        SigIntMixin.term_sig_handler(self, sig, frame)
    
    def sig_handler(self, sig):
        if sig in self.term_sigs:
            print "Terminate signal received through pipe"
            self.terminate_process()
    
    def setup_handlers(self):
        if self.use_pipe:
            self.pipe = SelfPipe(self.sig_handler)
            self.pipe.track_signals(*self.term_sigs)
        else:
            self.setup_term_sig_handler()
    
    def setup_child(self, child):
        pass
    
    def cleanup_request(self, request, client_address):
        pass
    
    def process_request(self, request, client_address):
        if self.wrapper_class:
            child = self.wrapper_class(TCPServer.process_request,
                                       self, request, client_address)
            self.setup_child(child)
            print "Starting child", child
            child.start()
            self.cleanup_request(request, client_address)
        else:
            TCPServer.process_request(self, request, client_address)
    
    def terminate_process(self):
        self.__shutdown_flag = True
    
    def serve_forever(self):
        # We won't be using the built-in shutdown mechanism of TCPServer
        # since it gives us no way to access it using the self-pipe trick
        # (at least, not without spawning a separate thread just to call
        # shutdown, to avoid deadlock!), so we override this; note that
        # we can't use the built-in handle_request method either, since it
        # also gives no way to select on the pipe
        
        while not self.__shutdown_flag:
            if self.test_race:
                # This demonstrates a race condition between receiving a
                # signal and starting the select call below; using a
                # standard signal handler, a Ctrl-C during the sleep
                # call below will not terminate the server; it will
                # simply interrupt the sleep call and go into the select
                # call, which will block until a request is received,
                # and *then* the program will terminate *after* handling
                # the request; the self-pipe trick removes this bug, a
                # Ctrl-C during the sleep will terminate the server
                # immediately, even if there is a request pending
                from time import sleep
                self.test_race = False
                print "Sleeping..."
                sleep(10)
            try:
                r, w, e = select.select(
                    [self.pipe, self] if self.pipe else [self],
                    [], []
                )
                if self.pipe in r:
                    self.pipe.receive_signals()
                    # go to top of loop to check shutdown flag now
                elif self in r:
                    self._handle_request_noblock()
            except (socket.error, select.error) as e:
                if e.args[0] in (EINTR, ERESTART):
                    print "System call interrupted"
                    continue
                else:
                    self.handle_error()


def server_class(opts, args):
    
    if opts.childtype == PROCESS:
        from plib.comm._processwrapper import ProcessWrapper
        from plib.comm.classes import SigChldMixin
        from plib.stdlib.decotools import wraps_class
        
        @wraps_class(DemoServer)
        class _Server(SigChldMixin, DemoServer):
            
            wrapper_class = ProcessWrapper
            
            def child_sig_handler(self, sig, frame=None):
                print "Child exit signal received by handler"
                SigChldMixin.child_sig_handler(self, sig, frame)
            
            def sig_handler(self, sig):
                if sig == signal.SIGCHLD:
                    print "Child exit signal received through pipe"
                    self.reap_children()
                else:
                    DemoServer.sig_handler(self, sig)
            
            def setup_handlers(self):
                DemoServer.setup_handlers(self)
                if hasattr(signal, 'SIGCHLD'):
                    if self.use_pipe:
                        self.pipe.track_signal(signal.SIGCHLD, reset=True)
                    else:
                        self.setup_child_sig_handler()
            
            def setup_child(self, child):
                self.track_child(child)
            
            def check_child(self, child):
                return child.check()
            
            def end_child(self, child):
                child.end()
            
            def cleanup_request(self, request, client_address):
                # Only the child process needs the request open now
                self.close_request(request)
    
    elif opts.childtype == THREAD:
        from plib.comm._threadwrapper import ThreadWrapper
        from plib.stdlib.decotools import wraps_class
        
        @wraps_class(DemoServer)
        class _Server(DemoServer):
            
            wrapper_class = ThreadWrapper
            
            def check_child(self, child):
                return child.check()
            
            def end_child(self, child):
                child.end()
    
    else:
        _Server = DemoServer
    
    return partial(_Server,
                   ("localhost", int(args.portnum)),
                   DemoRequestHandler,
                   opts.use_pipe, opts.test_race)


server_optlist = (
    ('-p', '--process', {
        'action': 'store_const', 'const': PROCESS,
        'dest': 'childtype', 'default': 0,
        'help': "fork child process for worker"
    }),
    ('-r', '--race', {
        'action': 'store_true',
        'dest': 'test_race', 'default': False,
        'help': "test for race condition between signal handler and select"
    }),
    ('-t', '--thread', {
        'action': 'store_const', 'const': THREAD,
        'dest': 'childtype', 'default': 0,
        'help': "start child thread for worker"
    }),
    ('-u', '--use-pipe', {
        'action': 'store_true',
        'dest': 'use_pipe', 'default': False,
        'help': "use the self-pipe trick"
    })
)

server_arglist = ["portnum"]


if __name__ == '__main__':
    from plib.stdlib.options import parse_options
    
    opts, args = parse_options(server_optlist, server_arglist)
    
    server_class(opts, args)().serve_forever()



#!/usr/bin/env python
"""
CLIENTSERVER.PY
Copyright (C) 2008-2015 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

A client program for the demonstration server that forks
the server itself.
"""

import sys
import os
import select
import signal
import socket
from errno import EINTR, ERESTART

QUIT = "Q"


def run_client(portnum, shutdown):
    while True:
        cmd = raw_input("Hit Enter to connect, or type Q to quit: ")
        if cmd.encode() == QUIT:
            # This is equivalent to sending a Ctrl-C to the server if you are
            # using the test-race functionality
            break
        
        else:
            # The server will print a message and close the connection immediately
            try:
                socket.create_connection(("localhost", portnum))
            except socket.error as e:
                print str(e)
            else:
                print "Connected ok"
    
    shutdown()


def run_clientserver(server_class, use_socketpair, portnum):
    if use_socketpair:
        from plib.comm._threadwrapper import ThreadWrapper
        from plib.comm.socketpair import fork_socketpair
        
        def sock_listener(sock):
            while True:
                try:
                    r, w, e = select.select([sock], [], [])
                    if sock in r:
                        cmd = sock.recv(4096)
                        if cmd == QUIT:
                            break
                except select.error as e:
                    if e.args[0] in (EINTR, ERESTART):
                        continue
                    else:
                        break
            
            # This signal gets handled in the main thread, so the
            # server will catch it and break out of serve_forever
            os.kill(os.getpid(), signal.SIGINT)
        
        def run_server(sock):
            listener = ThreadWrapper(sock_listener, sock)
            listener.start()
            server_class().serve_forever()
            listener.wait()  # should shut itself down after os.kill
        
        client_sock, server = fork_socketpair(run_server)
        
        def shutdown():
            client_sock.sendall(QUIT)
            server.wait()
    
    else:
        from plib.comm._processwrapper import ProcessWrapper
        from plib.comm.forkserver import fork_server
        
        ProcessWrapper.term_sig = signal.SIGINT  # put this into API?
        server = fork_server(server_class)
        shutdown = server.end
    
    run_client(portnum, shutdown)


if __name__ == '__main__':
    from plib.stdlib.options import parse_options
    
    from server import server_class, server_optlist, server_arglist
    
    optlist = server_optlist + (
        ('-s', '--use-socketpair', {
            'action': 'store_true',
            'dest': 'use_socketpair', 'default': False,
            'help': "use socketpair to communicate with forked server"
        }),
    )
    
    arglist = server_arglist
    
    opts, args = parse_options(optlist, arglist)
    
    run_clientserver(server_class(opts, args), opts.use_socketpair, args.portnum)



#!/usr/bin/env python
"""
Module FORKWAIT -- Specialized Forking Function
Sub-Package COMM of Package PLIB -- General Python Utilities
Copyright (C) 2008-2015 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the ``fork_wait`` function, which
forks a subprocess and then waits until the subprocess
has started before continuing.

Note that on Windows the ``multiprocessing`` module is used,
which is only available in Python 2.6 and later.
"""

from ._processwrapper import ProcessWrapper
from ._waitwrapper import wait_wrapper


def fork_wait(start_fn, run_fn):
    return wait_wrapper(start_fn, run_fn, ProcessWrapper)



#!/usr/bin/env python
"""
Module SigIntMixin
Sub-Package COMM.CLASSES of Package PLIB
Copyright (C) 2008-2015 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the ``SigIntMixin`` class. This is a mixin
class for servers that can be closed by Ctrl-C (or by other
termination signals if the ``term_sigs`` class field is
changed in a subclass), but which do not need any other complex
functionality.
"""

import signal


class SigIntMixin(object):
    """Mixin class for processes to do controlled shutdown on Ctrl-C.
    
    The signals to be trapped and treated as "termination" signals
    can be changed by overriding the ``term_sigs`` class field.
    
    Note that this is not, strictly speaking, new functionality;
    Python by default traps Ctrl-C and raises KeyboardInterrupt,
    which by default causes a program exit; but that does not allow
    for the program to do any shutdown processing unless it
    explicitly traps the exception, which does not always fit in
    with the program's structure. This class traps the signal and
    integrates it into the program's general signal handler. The
    Python default also prints a messy traceback, which this class
    eliminates.
    """
    
    term_sigs = [signal.SIGINT]
    
    def setup_term_sig_handler(self):
        """Set up the termination signal handler.
        """
        for sig in self.term_sigs:
            signal.signal(sig, self.term_sig_handler)
    
    def term_sig_handler(self, sig, frame=None):
        """Shut down the process when a terminate signal is received.
        """
        self.terminate_process()
    
    def terminate_process(self):
        """Terminate the process.
        
        This is intended to be overridden in derived classes to
        implement the appropriate shutdown mechanism. For an example,
        see the ``SigIntServerMixin`` class in ``plib.io.mixins``,
        which is customized for use with PLIB servers.
        """
        pass



#!/usr/bin/env python
"""
Module SIGTOOLS -- Signal Handling Utilities
Sub-Package COMM of Package PLIB
Copyright (C) 2008-2015 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains utilities for use in handling
signals.
"""

import os
import atexit
import signal
from functools import partial

from plib.stdlib.coll import fifo


# First the os-specific code

if os.name == 'posix':
    # We're on a Unix-type system and can use os.pipe with select
    
    from plib.stdlib.fdtools import set_nonblocking
    
    
    _read_fd = None
    _write_fd = None
    
    
    def _cleanup_pipe(*fds):
        for fd in fds:
            os.close(fd)
    
    
    def _setup_pipe():
        global _read_fd, _write_fd
        if _read_fd is None:
            _read_fd, _write_fd = os.pipe()
            for fd in (_read_fd, _write_fd):
                set_nonblocking(fd)
            signal.set_wakeup_fd(_write_fd)
            atexit.register(partial(_cleanup_pipe, _read_fd, _write_fd))
    
    
    def selfpipe_fd():
        _setup_pipe()
        return _read_fd
    
    
    def _flush_pipe():
        # The pipe will contain all \x00 bytes, so there's no
        # use retaining them; we just need to allow for either
        # an os.error or a zero-byte read signaling that the
        # pipe is empty
        while True:
            try:
                b = os.read(_read_fd, 1)
            except os.error:
                break
            else:
                if not b:
                    break
    
    
    def send_signal(sig):
        os.kill(os.getpid(), sig)


else:
    # We're on Windows and have to use an (emulated) socketpair to
    # emulate a select-able pipe
    
    from plib.comm._threadwrapper import ThreadWrapper
    from plib.comm._socketwrapper import socketpair_wrapper
    
    
    _read_sock = None
    _write_sock = None
    
    
    def _cleanup_pipe(*socks):
        for sock in socks:
            sock.close()
    
    
    def _pipe(wsock):
        global _write_sock
        wsock.setblocking(0)
        _write_sock = wsock
    
    
    def _setup_pipe():
        global _read_sock
        if _read_sock is None:
            _read_sock, child = socketpair_wrapper(
                _pipe, ThreadWrapper)
            _read_sock.setblocking(0)
            child.wait()  # _write_sock will be filled in when this returns
            signal.set_wakeup_fd(_write_sock.fileno())
            atexit.register(partial(_cleanup_pipe, _read_sock, _write_sock))
    
    
    def selfpipe_fd():
        _setup_pipe()
        return _read_sock.fileno()
    
    
    def _flush_pipe():
        # The pipe will contain all \x00 bytes, so there's no
        # use retaining them; we just need to allow for either
        # a socket.error or a zero-byte read signaling that the
        # pipe is empty
        while True:
            try:
                b = _read_sock.recv(1)
            except socket.error:
                break
            else:
                if not b:
                    break
    
    
    def send_signal(sig):
        _sigs_received.append(sig)
        try:
            _write_sock.send('\x00')
        except socket.error:
            # If the pipe is already full we are fine anyway
            pass


# Now the common code

_reset_signals = set()

_sigs_received = fifo()


def _sighandler(sig, frame):
    _sigs_received.append(sig)
    if sig in _reset_signals:
        # These signals have to have the handler reset every time
        signal.signal(sig, _sighandler)


_old_handlers = {}


def _setup_sig(sig, reset=False):
    if sig not in _old_handlers:
        _setup_pipe()
        _old_handlers[sig] = signal.signal(sig, _sighandler)
        if reset:
            _reset_signals.add(sig)


def track_signal(sig, reset=False):
    _setup_sig(sig, reset)
    return _old_handlers[sig]


def track_signals(*sigs):
    for sig in sigs:
        _setup_sig(sig)
    return tuple(_old_handlers[sig] for sig in sigs)


def _cleanup_sig(sig):
    if sig in _old_handlers:
        signal.signal(sig, _old_handlers[sig])
        del _old_handlers[sig]
        if sig in _reset_signals:
            _reset_signals.discard(sig)


def untrack_signal(sig):
    _cleanup_sig(sig)


def untrack_signals(*sigs):
    for sig in sigs:
        _cleanup_sig(sig)


def signals_received():
    # Flush the pipe first, then check signals received, to avoid
    # a race if a signal arrives after exiting the while loop below
    _flush_pipe()
    while _sigs_received:
        yield _sigs_received.nextitem()



#!/usr/bin/env python
"""
Module THREADWRAPPER -- Child Thread Wrapper Object
Sub-Package COMM of Package PLIB -- General Python Utilities
Copyright (C) 2008-2015 by Peter A. Donis

This module contains the ``ThreadWrapper`` class, which provides
a portable interface to child threads and their exit status.
This class is used by the specialized threading functions in this
library.
"""

import threading

from ._childwrapper import ChildWrapper


class ChildThread(threading.Thread):
    # Fortunately we don't have to go through the same gymnastics
    # as for child processes (see the _processwrapper module)
    
    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs=None, verbose=None):
        # We save this info so run can see it below (the base class
        # stores them as private vars)
        self._target = target
        self._args = args
        self._kwargs = kwargs
        threading.Thread.__init__(self, group=group, name=name,
                                  verbose=verbose)
    
    def run(self):
        # And now we need to actually return the exit code
        # (at least this part is easily implemented...)
        try:
            if self._target:
                self.exitcode = self._target(*self._args, **self._kwargs)
        finally:
            # Avoid a refcycle if the thread is running a function with
            # an argument that has a member that points to the thread.
            del self._target, self._args, self._kwargs


class ThreadWrapper(ChildWrapper):
    
    thread_class = ChildThread
    
    thread = None
    
    def start(self):
        if self.thread is None:
            self.thread = self.thread_class(
                target=self._target, args=self._args, kwargs=self._kwargs)
            self.thread.daemon = self.shutdown_with_parent
            self.thread.start()
        else:
            raise RuntimeError("Thread already started.")
    
    def stop(self):
        # No way to stop a thread by default; this method can be overridden
        # to allow interaction with the thread via events or other means (in
        # such cases the thread_class class field can also be overridden to
        # use a subclass of ChildThread)
        pass
    
    def check(self):
        if self.thread is None:
            return True
        if not self.thread.is_alive():
            self._exitcode = self.thread.exitcode
            self.thread = None
            return True
        return False
    
    def wait(self):
        self.thread.join()
        self._exitcode = self.thread.exitcode
        self.thread = None



#!/usr/bin/env python
"""
Module SOCKETPAIR -- Socket Pair Forking Function
Sub-Package COMM of Package PLIB -- General Python Utilities
Copyright (C) 2008-2015 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the fork_socketpair function, which
forks a subprocess that communicates with its parent via
a socket pair.

Note that on Windows the ``multiprocessing`` module is used,
which is only available in Python 2.6 and later.
"""

from ._processwrapper import ProcessWrapper
from ._socketwrapper import socketpair_wrapper


def fork_socketpair(fn):
    return socketpair_wrapper(fn, ProcessWrapper)
